#include <iostream>
#include <v8.h>
#include <node.h>
#include <node_buffer.h>
#include <stdlib.h>
#include <string.h>
//#include "miniz.c"
#include <zlib.h>
#include "easyzlib.h"

#define MIN_BYTES 32

using namespace std;
using namespace v8;
using namespace node;

/* easy zlib functions
   if destination buffer is not large enough, destination length set to required length
*/
int ezcompress( unsigned char* pDest, long* pnDestLen, const unsigned char* pSrc, long nSrcLen, int level, int windowBits, int memLevel, int strategy)
{
    z_stream stream;
    int err;

	int nExtraChunks;
	uInt destlen;

    stream.next_in = (Bytef*)pSrc;
    stream.avail_in = (uInt)nSrcLen;

#ifdef MAXSEG_64K
    /* Check for source > 64K on 16-bit machine: */
    if ((uLong)stream.avail_in != nSrcLen) return Z_BUF_ERROR;
#endif
	
	destlen = (uInt)*pnDestLen;
	if ((uLong)destlen != (uLong)*pnDestLen) return Z_BUF_ERROR;

    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;
    stream.opaque = (voidpf)0;

    err = deflateInit2(&stream, level, Z_DEFLATED, windowBits, memLevel, strategy);
    if (err != Z_OK) return err;

	nExtraChunks = 0;
    do {
		stream.next_out = pDest;
		stream.avail_out = destlen;
        err = deflate(&stream, Z_FINISH);
		if (err == Z_STREAM_END )
			break;
		if (err != Z_OK) {
			deflateEnd(&stream);
			return err;
		}
		nExtraChunks += 1;
    } while (stream.avail_out == 0);

    *pnDestLen = stream.total_out;

    err = deflateEnd(&stream);
    if (err != Z_OK) return err;

    return nExtraChunks ? Z_BUF_ERROR : Z_OK;
}

int ezuncompress( unsigned char* pDest, long* pnDestLen, const unsigned char* pSrc, long nSrcLen, int windowBits)
{
    z_stream stream;
    int err;

	int nExtraChunks;
	uInt destlen;

    stream.next_in = (Bytef*)pSrc;
    stream.avail_in = (uInt)nSrcLen;

#ifdef MAXSEG_64K
    /* Check for source > 64K on 16-bit machine: */
    if ((uLong)stream.avail_in != (uLong)nSrcLen) return Z_BUF_ERROR;
#endif

	destlen = (uInt)*pnDestLen;
	if ((uLong)destlen != (uLong)*pnDestLen) return Z_BUF_ERROR;

    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;

    err = inflateInit2(&stream, windowBits);
    if (err != Z_OK) return err;

	nExtraChunks = 0;
    do {
		stream.next_out = pDest;
		stream.avail_out = destlen;
        err = inflate(&stream, Z_FINISH);
		if (err == Z_STREAM_END )
			break;
		if (err == Z_NEED_DICT || (err == Z_BUF_ERROR && stream.avail_in == 0))
			err = Z_DATA_ERROR;
		if (err != Z_BUF_ERROR) {
			inflateEnd(&stream);
			return err;
		}
		nExtraChunks += 1;
    } while (stream.avail_out == 0);

    *pnDestLen = stream.total_out;

    err = inflateEnd(&stream);
    if (err != Z_OK) return err;

    return nExtraChunks ? Z_BUF_ERROR : Z_OK;
}
//End of C/C++ implementation


#define ThrowNodeError(what) {\
	return ThrowException(Exception::Error(String::New(what))); \
}

enum Format{
	ZLIB = 0,
	GZIP = 1,
	DEFLATE = 2 
	
};

Handle<Value> Common(const Arguments &args, bool compress, Format f = ZLIB)
{
	HandleScope scope;
	
	if (args.Length() < 1) { 
		ThrowNodeError("Argument 0 is not string or buffer");
	}
	
	ezbuffer bufSrc;
	
	if (args[0]->IsString()) {
		String::AsciiValue string(args[0]->ToString());
		
		if (string.length() <= MIN_BYTES) {
			return scope.Close(args[0]);
		}
		bufSrc.Alloc(string.length());
		bufSrc.nSize = bufSrc.nLen = string.length();
		memcpy(bufSrc.pBuf, *string, bufSrc.nLen);
				
	} else if (Buffer::HasInstance(args[0])) {
		
		Local<Object> bufferIn = args[0]->ToObject();
				
		if (Buffer::Length(bufferIn) <= MIN_BYTES) {
			return scope.Close(args[0]);
		}
		bufSrc.Alloc(Buffer::Length(bufferIn));
		bufSrc.nSize = bufSrc.nLen = Buffer::Length(bufferIn);
		memcpy(bufSrc.pBuf, Buffer::Data(bufferIn), bufSrc.nLen);
		
	} else {
		ThrowNodeError("Argument 0 is not string or buffer");
	}
	
	int compressionLevel = Z_DEFAULT_COMPRESSION,
		memLevel = 8,
		strategy = Z_DEFAULT_STRATEGY,
		windowBits = MAX_WBITS,
		number;
	
	if (args.Length() > 1) { 
		if(compress && args[0]->IsObject()){
			Local<Object> options = args[0]->ToObject();
						
			if(options->Has(String::New("compressionLevel"))){
				number = options->Get(String::New("compressionLevel"))->IntegerValue();
				if (number > 0 && number <= 9) {
					compressionLevel = number;
				}
			}
			
			if(options->Has(String::New("windowBits"))){
				number = options->Get(String::New("windowBits"))->IntegerValue();
				if (number >= 8 && number <= MAX_WBITS) {
					windowBits = number;
				}
			}
			
			if(options->Has(String::New("memLevel"))){
				number = options->Get(String::New("memLevel"))->IntegerValue();
				if (number >= 1 && number <= 9) {
					memLevel = number;
				}
			}
			
			if(options->Has(String::New("strategy"))){
				number = options->Get(String::New("strategy"))->IntegerValue();
				if (number >= 0 && number <= 4) {
					strategy = number;
				}
			}
		}else if(!compress && args[0]->IsNumber()){
			number = args[0]->IntegerValue();
			if (number >= 8 && number <= MAX_WBITS) {
				windowBits = number;
			}
		}else{
			ThrowNodeError("Invalid options");
		}
	
	}
	
		
	switch(strategy){
	
		case 1:
			strategy = Z_FILTERED;
		break;
		
		case 2:
			strategy = Z_HUFFMAN_ONLY;
		break;
		
		case 3:
			strategy = Z_RLE;
		break;
		
		case 4:
			strategy = Z_FIXED;
		break;
		
		default:
			strategy = Z_DEFAULT_STRATEGY;
		break;
	}

	
	
	switch(f){
		
		case ZLIB:
		break;
		
		case GZIP:
			windowBits += 16;
		break;
		
		case DEFLATE:
			windowBits = -windowBits;
		break;
		
	}
	

	
	ezbuffer bufDest(compress ? bufSrc.nLen : bufSrc.nLen << 2);

	int ret = compress ? ezcompress(bufDest, bufSrc, compressionLevel, windowBits, memLevel, strategy) : ezuncompress(bufDest, bufSrc, windowBits);
	
	if(ret != Z_OK)
    {
        switch(ret)
        {
            case EZ_MEM_ERROR:
				ThrowNodeError("Memory error!"); 

            case EZ_BUF_ERROR:
				ThrowNodeError("Buffer error!");
			
			case EZ_DATA_ERROR:
				ThrowNodeError("Data error!");
				
			case EZ_STREAM_ERROR:
				ThrowNodeError("Stream error!");
			
            default:
                ThrowNodeError("Unknown error: ");
        }

    }


	Buffer *BufferOut = Buffer::New((char *)bufDest.pBuf, bufDest.nLen);

	return scope.Close(BufferOut->handle_);
}



Handle<Value> gzipcompress(const Arguments &args)
{
	return Common(args, true, GZIP);
}

Handle<Value> gzipuncompress(const Arguments &args)
{
	return Common(args, false, GZIP);
}

Handle<Value> zlibcompress(const Arguments &args)
{
	return Common(args, true, ZLIB);
}

Handle<Value> zlibuncompress(const Arguments &args)
{
	return Common(args, false, ZLIB);
}

Handle<Value> deflatecompress(const Arguments &args)
{
	return Common(args, true, DEFLATE);
}

Handle<Value> deflateuncompress(const Arguments &args)
{
	return Common(args, false, DEFLATE);
}

extern "C" void
init (Handle<Object> target) {

	NODE_SET_METHOD(target, "gzipcompress", gzipcompress);
	NODE_SET_METHOD(target, "gzipuncompress", gzipuncompress);
	NODE_SET_METHOD(target, "zlibcompress", zlibcompress);
	NODE_SET_METHOD(target, "zlibuncompress", zlibuncompress);
	NODE_SET_METHOD(target, "deflatecompress", deflatecompress);
	NODE_SET_METHOD(target, "deflateuncompress", deflateuncompress);
}

